unit zlib_interface;
{<
  zlib_interface contains interface utility for easy using of zlib or bzip2 functionality.
  _rosi.inc contain switch for zlib or bzip2 usage.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Apr 2014)
  @lastmod(Oct 2014)
}

interface

{$INCLUDE _delphi_ver.inc}
{$INCLUDE _rosi.inc}
uses Sysutils, Classes;

function CompressStream(Source, Dest: TStream): Int64;
function DeCompressStream(Source, Dest: TStream): Int64;

function CompressFile(SourceFile, DestFile: string): Int64;
function DeCompressFile(SourceFile, DestFile: string): Int64;

function CompressString(SourceStr: string; SaveAsFile: string = ''): string;
function DeCompressString(SourceStr: string; LoadFromFile: string = ''): string;

implementation

{$IFDEF bzip2}
uses bzip2_main;

function MyCompStreamCreate(Dest: TStream): TStream;
begin
  Result:=TBZCompressionStream.Create(bs9, Dest);
end;

function MyDecompStreamCreate(Source: TStream): TStream;
begin
  Result:=TBZDecompressionStream.Create(Source);
end;
{$ELSE}
uses zlib;

function MyCompStreamCreate(Dest: TStream): TStream;
begin
  Result:=TZCompressionStream.Create(Dest);
end;

function MyDecompStreamCreate(Source: TStream): TStream;
begin
  Result:=TZDecompressionStream.Create(Source);
end;
{$ENDIF}

function CompressStream(Source, Dest: TStream): Int64;
var
  Comp: TStream;
begin
  Comp:=MyCompStreamCreate(Dest);
  try
    Comp.CopyFrom(Source, 0);
  finally
    Comp.Free;
  end;
  Result:=Dest.Size;
end;

function DeCompressStream(Source, Dest: TStream): Int64;
const
  BufferSize = 65536;
var
  Comp: TStream;
  Count: Integer;
  Buffer: TBytes;
begin
  Comp:=MyDecompStreamCreate(Source);
  try
    SetLength(Buffer, BufferSize);
    repeat
      Count:=Comp.Read(Buffer[0], BufferSize);
      if Count <> 0 then
        Dest.WriteBuffer(Buffer[0], Count)
      else
        Break;
    until false;
    Result:=Dest.Size;
  finally
    Comp.Free;
  end;
end;

function CompressFile(SourceFile, DestFile: string): Int64;
var
  Source, Dest: TFileStream;
begin
  Source:=TFileStream.Create(SourceFile, fmOpenRead + fmShareDenyNone);
  try
    Dest:=TFileStream.Create(DestFile, fmCreate);
    try
      Result:=CompressStream(Source, Dest);
    finally
      Dest.Free;
    end;
  finally
    Source.Free;
  end;
end;

function DeCompressFile(SourceFile, DestFile: string): Int64;
var
  Source, Dest: TFileStream;
begin
  Source:=TFileStream.Create(SourceFile, fmOpenRead + fmShareDenyWrite);
  try
    Dest:=TFileStream.Create(DestFile, fmCreate);
    try
      Result:=DeCompressStream(Source, Dest);
    finally
      Dest.Free;
    end;
  finally
    Source.Free;
  end;
end;

function CompressString(SourceStr: string; SaveAsFile: string = ''): string;
var
  Source, Dest: TStringStream;
begin
  Source:=TStringStream.Create(SourceStr);
  try
    Dest:=TStringStream.Create('');
    try
      CompressStream(Source, Dest);
      Dest.Seek(0, soFromBeginning);
      Result:=Dest.DataString;
      if SaveAsFile <> '' then
        Dest.SaveToFile(SaveAsFile);
    finally
      Dest.Free;
    end;
  finally
    Source.Free;
  end;
end;

function DeCompressString(SourceStr: string; LoadFromFile: string = ''): string;
var
  Source, Dest: TStringStream;
begin
  Source:=TStringStream.Create(SourceStr);
  if LoadFromFile <> '' then
    Source.LoadFromFile(LoadFromFile);
  try
    Dest:=TStringStream.Create;
    try
      DeCompressStream(Source, Dest);
      Dest.Seek(0, soFromBeginning);
      Result:=Dest.DataString;
    finally
      Dest.Free;
    end;
  finally
    Source.Free;
  end;
end;

end.
