unit ADOConWizard;

{<
  ADOConWizard is non visual component which provides all functions
  for connecting to database using TADOConnection and UDL files.
  ADOConWizard can use more external UDL files for generating of connection
  string, so one application can work under several systems
  (which uses different connection string).
}

{
  Date: Dec 2005
  Updated version compatible with Delphi2010+
  Last update March 2011

  Author: Rosi (http://www.rosinsky.cz/delphi)
  Full functional demo
}

interface

uses
  Windows, Messages, SysUtils, Classes, Controls,
  ADODB, Dialogs, Forms;

type
  TProviderType = (ptNone, ptMSDA, ptJET, ptACE, ptSQLOLEDB);
  TConnectionEvent = (AfterConnecting, AfterLogin, AfterOpening);
  TConnectionEvents = set of TConnectionEvent;
  { Defines events when error dialog can be shown }
  TOnInfo = procedure(Sender: TObject; Info: string) of object;
  { Procedure is called for each step in connection, action is described by
    "Info" string }
  TOnGetDBFilePassword = procedure(Sender: TObject; var UserID, Password: string) of object;
  { Procedure is called before connecting to obtain User and Password to file }
  TOnLoginUser = procedure(Sender: TObject; var Logon: boolean) of object;
  { Procedure is called after connecting to DB before opening of tables,
    so it allows to make user logon;
    if parameter Logon is changed to false, DB is disconnected }
  TOnOpenTables = procedure(Sender: TObject; var Opened: boolean) of object;
  { Procedure is called after user logon, so it allows to open all tables;
    if parameter Opened is changed to false, DB is disconnected }

  TADOConWizard = class(TComponent)
  private
    OpenDlg: TOpenDialog;
    WasCon: boolean;
    ProviderType: TProviderType;
    FADOConnection: TADOConnection;
    FDefaultDBFile: string;
    FUDLFilePath: string;
    FUDLFileMask: string;
    FErrDialogShow: TConnectionEvents;
    FLogDetails: boolean;
    FOnInfo: TOnInfo;
    FOnGetDBFilePassword: TOnGetDBFilePassword;
    FOnLoginUser: TOnLoginUser;
    FOnOpenTables: TOnOpenTables;
    procedure Info(s: string);
    procedure FindUDLFiles(List: TStringList);
    function GetProviderType(ConStr: string): TProviderType;
    function FindConStr(UDLFile: string): string;
    function FormatConStr(ConStr, FileName: string): string;
    function GetFileName(ConStr: string): string;
    function GetConnectedFile: string;
    function OpenAll_Sel(FileName: string): boolean;
    function OpenAll(SelectedFile: string): boolean;
    function CheckPsw: boolean;
    function OpenTables: boolean;
  protected
    procedure SetADOConnection(Value: TADOConnection);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    idProviderStr, idEndStr: string;
    idProvider, idDBFile, idDefaultDir, idPsw: array [TProviderType] of string;
    { Format strings for generating of connection string }
    LogText: TStringList;
    { List of all info strings - log file }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    function ConnectToStr(ConStr: string): boolean;
    { Tries to connect using connection string "ConStr", returns the
      result (true=connected) }
    function ConnectToFile(DBFileName: string; UseUDLFiles: boolean): boolean;
    { Tries to connect to DBFileName using current connection string
      and/or using UDL files if "UseUDLFiles" is true, returns the result }
    function Connect: boolean;
    { Tries to connect using current connection string
      and/or to current DBFile using UDL files, returns the result }
    procedure Disconnect;
    { Disconnect connection }
    function OpenAll_LastFile: boolean;
    { Tries to connect using current connection string
      and/or to current DBFile using UDL files, if it is empty, try to
      connect to "DefaultDBFile", returns the result;
      function provides error and select file dialog }
    function OpenAll_SelectFile: boolean;
    { Shows select file dialog and try to connect to selected file
      using UDL files, returns the result;
      function provides error and select file dialog }
    function OpenAll_ToFile(FileName: string): boolean;
    { Tries to connect to "FileName" file using UDL files,
      returns the result;
      function provides error and select file dialog }
    function CreateUDLFile: boolean;
    { Creates new UDL file and opens it for editing,
      returns the result }
    function SaveLog(var LogFileName: string): boolean;
    { Saves "LogText" to "LogFileName", returns the result }
  published
    { Published declarations }
    property ADOConnection: TADOConnection read FADOConnection write SetADOConnection;
    { Link to TADOConnection }
    property DefaultDBFile: string read FDefaultDBFile write FDefaultDBFile;
    property UDLFilePath: string read FUDLFilePath write FUDLFilePath;
    property UDLFileMask: string read FUDLFileMask write FUDLFileMask;
    { Path with UDL files }
    property ConnectedFile: string read GetConnectedFile;
    property LogDetails: boolean read FLogDetails write FLogDetails default false;
    property ErrDialogShow: TConnectionEvents read FErrDialogShow write FErrDialogShow default[AfterConnecting, AfterLogin, AfterOpening];
    { Defines after which events Error dialog could be shown }
    property OnInfo: TOnInfo read FOnInfo write FOnInfo;
    property OnGetDBFilePassword: TOnGetDBFilePassword read FOnGetDBFilePassword write FOnGetDBFilePassword;
    property OnLoginUser: TOnLoginUser read FOnLoginUser write FOnLoginUser;
    property OnOpenTables: TOnOpenTables read FOnOpenTables write FOnOpenTables;
  end;

procedure Register;

implementation

{$R *.res}
uses rstring, rtool, rdlg, ADOConWizard_Form;

procedure Register;
begin
  RegisterComponents('Rosi DB', [TADOConWizard]);
end;

// ======================================================================
// Public actions
constructor TADOConWizard.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  LogText:=TStringList.Create;
  OpenDlg:=TOpenDialog.Create(Self);
  OpenDlg.Filter:=sFileTypeMDB + '|' + sFileTypeAll;
  OpenDlg.InitialDir:=GetAppPath;

  idProviderStr:='PROVIDER=';
  idEndStr:=';';
  idProvider[ptMSDA]:='PROVIDER=MSDASQL';
  idProvider[ptJET]:='PROVIDER=MICROSOFT.JET.OLEDB';
  idProvider[ptACE]:='PROVIDER=MICROSOFT.ACE.OLEDB';
  idProvider[ptSQLOLEDB]:='PROVIDER=SQLOLEDB';
  idDBFile[ptMSDA]:='DBQ=|SOURCEDB=';
  idDBFile[ptJET]:='DATA SOURCE=';
  idDBFile[ptACE]:='DATA SOURCE=';
  idDBFile[ptSQLOLEDB]:='INITIAL CATALOG=';
  idDefaultDir[ptMSDA]:='DEFAULTDIR=';
  idDefaultDir[ptJET]:='';
  idDefaultDir[ptACE]:='';
  idPsw[ptMSDA]:='PASSWORD=';
  idPsw[ptJET]:='JET OLEDB:DATABASE PASSWORD=';
  idPsw[ptACE]:='JET OLEDB:DATABASE PASSWORD=';
  idPsw[ptSQLOLEDB]:='PASSWORD=';
  ProviderType:=ptNone;

  WasCon:=false;
  FErrDialogShow:=[AfterConnecting, AfterLogin, AfterOpening];
end;

destructor TADOConWizard.Destroy;
begin
  LogText.Free;
  OpenDlg.Free;
  inherited Destroy;
end;

procedure TADOConWizard.Loaded;
begin
  inherited Loaded;
  if (csDesigning in ComponentState) then
    Exit;
  if FUDLFilePath = '' then
    FUDLFilePath:=GetAppPath;
end;

function TADOConWizard.ConnectToStr(ConStr: string): boolean;
var
  user, psw: string;
  oldCursor: TCursor;
begin
  Result:=false;
  if (FADOConnection = nil) then
    Exit;
  if ConStr = '' then
    Exit;
  oldCursor:=WCursor;
  try
    if LogDetails then
    begin
      Info('Try connection string: ');
      Info(ConStr);
    end;
    GetProviderType(ConStr);
    Info('DB FILE: ' + GetFileName(ConStr));
    // Try connect
    FADOConnection.Close;
    // Add password
    user:='';
    psw:='';
    if Assigned(FOnGetDBFilePassword) then
      FOnGetDBFilePassword(Self, user, psw);
    ConStr:=ReplaceStrFromTo(idPsw[ProviderType], idEndStr, '', ConStr, true, false, false);
    ConStr:=ConStr + idEndStr + idPsw[ProviderType] + psw + idEndStr;
    // Connect
    FADOConnection.ConnectionString:=ConStr;
    FADOConnection.Open;
    Result:=true;
    Info('CONNECTED. ');
  except
    on E: Exception do
      Info('ADO ERROR: ' + E.Message);
  end;
  Info('***');
  ChCursor(oldCursor);
end;

function TADOConWizard.ConnectToFile(DBFileName: string; UseUDLFiles: boolean): boolean;
var
  FList: TStringList;
  i: integer;
  ConStr: string;
begin
  Result:=false;
  if FADOConnection = nil then
    Exit;
  if DBFileName = '' then
    Exit;
  ConStr:=FADOConnection.ConnectionString;
  if ConStr <> '' then
    Result:=ConnectToStr(FormatConStr(ConStr, DBFileName));
  if Result then
    Exit;
  if not UseUDLFiles then
    Exit;
  FList:=TStringList.Create;
  try
    FindUDLFiles(FList);
    for i:=0 to FList.Count - 1 do
    begin
      if LogDetails then
        Info('Open UDL File: ' + FList[i]);
      ConStr:=FormatConStr(FindConStr(FList[i]), DBFileName);
      Result:=ConnectToStr(ConStr);
      if Result then
        Break;
    end;
  finally
    FList.Free;
  end;
end;

function TADOConWizard.Connect: boolean;
var
  ConStr: string;
begin
  ConStr:=FADOConnection.ConnectionString;
  if not ConnectToStr(ConStr) then
    ConnectToFile(GetFileName(ConStr), true);
  Result:=FADOConnection.Connected;
end;

procedure TADOConWizard.Disconnect;
begin
  FADOConnection.Close;
  if LogDetails then
    Info('DISCONNECTED. ');
end;

function TADOConWizard.OpenAll_LastFile: boolean;
begin
  Result:=OpenAll('');
end;

function TADOConWizard.OpenAll_SelectFile: boolean;
begin
  Result:=OpenAll_Sel('*');
end;

function TADOConWizard.OpenAll_ToFile(FileName: string): boolean;
begin
  Result:=OpenAll_Sel(FileName);
end;

function TADOConWizard.CreateUDLFile: boolean;
var
  a: integer;
  s: string;
  F: File;
begin
  Result:=false;
  // Find free UDL file
  a:=1;
  repeat
    s:=SetEndOfPath(FUDLFilePath) + Format('New%d.udl', [a]);
    Inc(a);
  until not FileExists(s);
  if LogDetails then
    Info('Make UDL File: ' + s);
  try
    AssignFile(F, s);
    Rewrite(F);
    CloseFile(F);
    RunFile(s, '', '', SW_SHOW);
    Sleep(3000);
    Result:=true;
  except
    on E: Exception do
      ShowErr(Format(sFileSaveErr, [s]), E, false);
  end;
end;

function TADOConWizard.SaveLog(var LogFileName: string): boolean;
begin
  if LogFileName = '' then
    LogFileName:=SetEndOfPath(FUDLFilePath) + ChangeFileExt(ExtractFileName(Application.ExeName), '.log');
  try
    LogText.SaveToFile(LogFileName);
    Result:=true;
  except
    Result:=false;
  end;
end;

// ======================================================================
// Private actions
procedure TADOConWizard.SetADOConnection(Value: TADOConnection);
begin
  FADOConnection:=Value;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TADOConWizard.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FADOConnection) then
    FADOConnection:=nil;
end;

procedure TADOConWizard.Info(s: string);
begin
  LogText.Add(s);
  if Assigned(FOnInfo) then
    FOnInfo(Self, s);
end;

procedure TADOConWizard.FindUDLFiles(List: TStringList);
var
  F: TSearchRec;
  Dir, Mask: string;
begin
  Dir:=SetEndOfPath(FUDLFilePath);
  if FUDLFileMask = '' then
    Mask:='*.udl'
  else
    Mask:=FUDLFileMask;
  List.Clear;
  if FindFirst(Dir + Mask, faAnyFile, F) = 0 then
    repeat
      List.Add(Dir + F.Name);
    until FindNext(F) <> 0 else Info('No UDL File in directory ' + Dir);
    FindClose(F);
  List.Sort;
end;

function TADOConWizard.FindConStr(UDLFile: string): string;
var
  i: integer;
  F: TextFile;
  s, r: string;
begin
  Result:='';
  AssignFile(F, UDLFile);
  try
    Reset(F);
    while not EOF(F) do
    begin
      ReadLn(F, s);
      // convert line from WideString to String
      r:='';
      for i:=1 to Length(s) do
        if (i mod 2 = 0) and (s[i] <> #0) then
          r:=r + s[i];
      if AnsiPosEx(idProviderStr, r, 1, false) = 1 then
        Result:=r;
    end;
    CloseFile(F);
  except
    on E: Exception do
      Info('UDL READ ERROR: ' + E.Message);
  end;
end;

function TADOConWizard.GetProviderType(ConStr: string): TProviderType;
begin
  ProviderType:=ptNone;
  if AnsiPosEx(idProvider[ptMSDA], ConStr, 1, false) = 1 then
    ProviderType:=ptMSDA;
  if AnsiPosEx(idProvider[ptJET], ConStr, 1, false) = 1 then
    ProviderType:=ptJET;
  if AnsiPosEx(idProvider[ptACE], ConStr, 1, false) = 1 then
    ProviderType:=ptACE;
  if AnsiPosEx(idProvider[ptSQLOLEDB], ConStr, 1, false) = 1 then
    ProviderType:=ptSQLOLEDB;

  Result:=ProviderType;
end;

function TADOConWizard.FormatConStr(ConStr, FileName: string): string;
var
  s, F: string;
  i: integer;
begin
  GetProviderType(ConStr);
  i:=0;
  repeat
    Inc(i);
    s:=FindStrPart(idDBFile[ProviderType], '|', i, false);
    // Replace DBQ Filename
    if (s <> '') and (AnsiPosEx(s, ConStr, 1, false) > 0) then
    begin
      // find filename in udl
      F:=FindStrFromTo(s, idEndStr, ConStr, false);
      F:=AnsiLowerCase(copy(F, Length(s) + 1, Length(F)));
      // use directory or filename?
      if ProviderType = ptSQLOLEDB then
        F:=FileName
      else
        if ExtractFileExt(F) = '' then
          F:=ExtractFilePath(FileName)
        else
          F:=FileName;
      Result:=ReplaceStrFromTo(s, idEndStr, s + F, ConStr, false, false, false);
    end;
  until s = '';
  i:=0;
  repeat
    Inc(i);
    s:=FindStrPart(idDefaultDir[ProviderType], '|', i, false);
    // Remove DefaultDir
    if s <> '' then
      Result:=ReplaceStrFromTo(s, idEndStr, '', Result, true, false, false);
  until s = '';
end;

function TADOConWizard.GetFileName(ConStr: string): string;
var
  s, F: string;
  i: integer;
begin
  i:=0;
  repeat
    Inc(i);
    s:=FindStrPart(idDBFile[ProviderType], '|', i, false);
    if (s <> '') and (AnsiPosEx(s, ConStr, 1, false) > 0) then
    begin
      F:=FindStrFromTo(s, idEndStr, ConStr, false);
      F:=AnsiLowerCase(copy(F, Length(s) + 1, Length(F) - Length(idEndStr)));
    end;
  until s = '';
  if (F = '') and Assigned(FADOConnection) then
    F:=FADOConnection.DefaultDatabase;
  Result:=F;
end;

function TADOConWizard.GetConnectedFile: string;
begin
  if Assigned(FADOConnection) and FADOConnection.Connected then
    Result:=GetFileName(FADOConnection.ConnectionString)
  else
    Result:='';
end;

function TADOConWizard.CheckPsw: boolean;
begin
  WasCon:=true;
  Result:=true;
  try
    if Assigned(FOnLoginUser) then
      FOnLoginUser(Self, Result);
  except
    Result:=false;
  end;
  if Result then
    Info('User Logon accepted.')
  else
  begin
    Info('User Logon NOT accepted.');
    Disconnect;
  end;
end;

function TADOConWizard.OpenTables: boolean;
begin
  Result:=true;
  try
    if Assigned(FOnOpenTables) then
      FOnOpenTables(Self, Result);
  except
    Result:=false;
  end;
  if Result then
  begin
    Info('Tables opened.');
    LogText.Clear;
  end
  else
  begin
    Info('Tables NOT opened.');
    Disconnect;
  end;
end;

function TADOConWizard.OpenAll_Sel(FileName: string): boolean;
var
  ConStr: string;
  WasConnected: boolean;
begin
  ConStr:=FADOConnection.ConnectionString;
  WasConnected:=FADOConnection.Connected;
  if not OpenAll(FileName) then
    if (not FADOConnection.Connected) and (ConStr <> '') and WasConnected then
    begin
      FADOConnection.ConnectionString:=ConStr;
      if DlgQ(sOpenLastFileQ) = mrYes then
        OpenAll('');
    end;
  Result:=FADOConnection.Connected;
end;

function TADOConWizard.OpenAll(SelectedFile: string): boolean;
  procedure ShowDialog;
  var
    s: string;
  begin
    s:=ExtractFileName(GetFileName(FADOConnection.ConnectionString));
    if ProviderType <> ptSQLOLEDB then
      s:=Format(sConnectDataErr, [s]) + #13 + sOpenOtherFileQ
    else
      s:=Format(sConnectDataErr, [s]) + #13 + sContinueQ;
    if ShowADOConDlg(Self, s, (not WasCon) and (SelectedFile <> '')) then
      if ProviderType <> ptSQLOLEDB then
        Result:=OpenAll('*')
      else
        Result:=OpenAll('');
  end;

begin
  if SelectedFile = '' then
  begin
    Disconnect;
    if FADOConnection.ConnectionString = '' then
      Result:=ConnectToFile(FDefaultDBFile, true)
    else
      Result:=Connect;
  end
  else
  begin
    if SelectedFile = '*' then
    begin
      Result:=OpenDlg.Execute;
      if not Result then
        Exit;
    end
    else
      OpenDlg.FileName:=SelectedFile;
    Disconnect;
    Result:=ConnectToFile(OpenDlg.FileName, true);
  end;
  if not Result then
  begin
    if AfterConnecting in FErrDialogShow then
      ShowDialog;
  end
  else
  begin
    Result:=CheckPsw;
    if not Result then
    begin
      if AfterLogin in FErrDialogShow then
        ShowDialog;
    end
    else
    begin
      Result:=OpenTables;
      if not Result then
      begin
        if AfterOpening in FErrDialogShow then
          ShowDialog;
      end;
    end;
  end;
end;

end.
